/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.  
 *
 * Copyright (c) 2010-2011, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

/** \file     TEncTop.cpp
    \brief    encoder class
*/

#include "../TLibCommon/CommonDef.h"
#include "TEncTop.h"

// ====================================================================================================================
// Constructor / destructor / create / destroy
// ====================================================================================================================

TEncTop::TEncTop()
{
  m_iPOCLast          = -1;
  m_iNumPicRcvd       =  0;
  m_uiNumAllPicCoded  =  0;
  m_pppcRDSbacCoder   =  NULL;
  m_pppcBinCoderCABAC =  NULL;
  m_cRDGoOnSbacCoder.init( &m_cRDGoOnBinCoderCABAC );
#if ENC_DEC_TRACE
  g_hTrace = fopen( "TraceEnc.txt", "wb" );
  g_bJustDoIt = g_bEncDecTraceDisable;
  g_nSymbolCounter = 0;
#endif

  m_iMaxRefPicNum     = 0;
}

TEncTop::~TEncTop()
{
#if ENC_DEC_TRACE
  fclose( g_hTrace );
#endif
}

Void TEncTop::create ()
{
  // initialize global variables
  initROM();
  
  // create processing unit classes
  m_cGOPEncoder.        create( getSourceWidth(), getSourceHeight(), g_uiMaxCUWidth, g_uiMaxCUHeight );
  m_cSliceEncoder.      create( getSourceWidth(), getSourceHeight(), g_uiMaxCUWidth, g_uiMaxCUHeight, g_uiMaxCUDepth );
  m_cCuEncoder.         create( g_uiMaxCUDepth, g_uiMaxCUWidth, g_uiMaxCUHeight );
#if MTK_SAO
  if (m_bUseSAO)
  {
    m_cEncSAO.create( getSourceWidth(), getSourceHeight(), g_uiMaxCUWidth, g_uiMaxCUHeight, g_uiMaxCUDepth );
    m_cEncSAO.createEncBuffer();
  }
#endif
  m_cAdaptiveLoopFilter.create( getSourceWidth(), getSourceHeight(), g_uiMaxCUWidth, g_uiMaxCUHeight, g_uiMaxCUDepth );
  m_cLoopFilter.        create( g_uiMaxCUDepth );

#if MQT_BA_RA && MQT_ALF_NPASS
  if(m_bUseALF)
  {
    m_cAdaptiveLoopFilter.setGOPSize( getGOPSize() );
    m_cAdaptiveLoopFilter.createAlfGlobalBuffers(m_iALFEncodePassReduction);
  }
#endif

  // if SBAC-based RD optimization is used
  if( m_bUseSBACRD )
  {
    m_pppcRDSbacCoder = new TEncSbac** [g_uiMaxCUDepth+1];
    m_pppcBinCoderCABAC = new TEncBinCABAC** [g_uiMaxCUDepth+1];
    
    for ( Int iDepth = 0; iDepth < g_uiMaxCUDepth+1; iDepth++ )
    {
      m_pppcRDSbacCoder[iDepth] = new TEncSbac* [CI_NUM];
      m_pppcBinCoderCABAC[iDepth] = new TEncBinCABAC* [CI_NUM];
      
      for (Int iCIIdx = 0; iCIIdx < CI_NUM; iCIIdx ++ )
      {
        m_pppcRDSbacCoder[iDepth][iCIIdx] = new TEncSbac;
        m_pppcBinCoderCABAC [iDepth][iCIIdx] = new TEncBinCABAC;
        m_pppcRDSbacCoder   [iDepth][iCIIdx]->init( m_pppcBinCoderCABAC [iDepth][iCIIdx] );
      }
    }
  }
}

Void TEncTop::destroy ()
{
#if MQT_BA_RA && MQT_ALF_NPASS
  if(m_bUseALF)
  {
    m_cAdaptiveLoopFilter.destroyAlfGlobalBuffers();
  }
#endif

  // destroy processing unit classes
  m_cGOPEncoder.        destroy();
  m_cSliceEncoder.      destroy();
  m_cCuEncoder.         destroy();
#if MTK_SAO
  if (m_cSPS.getUseSAO())
  {
    m_cEncSAO.destroy();
    m_cEncSAO.destoryEncBuffer();
  }
#endif
  m_cAdaptiveLoopFilter.destroy();
  m_cLoopFilter.        destroy();
  m_cBDS.               destroy();
  
  // SBAC RD
  if( m_bUseSBACRD )
  {
    Int iDepth;
    for ( iDepth = 0; iDepth < g_uiMaxCUDepth+1; iDepth++ )
    {
      for (Int iCIIdx = 0; iCIIdx < CI_NUM; iCIIdx ++ )
      {
        delete m_pppcRDSbacCoder[iDepth][iCIIdx];
        delete m_pppcBinCoderCABAC[iDepth][iCIIdx];
      }
    }
    
    for ( iDepth = 0; iDepth < g_uiMaxCUDepth+1; iDepth++ )
    {
      delete [] m_pppcRDSbacCoder[iDepth];
      delete [] m_pppcBinCoderCABAC[iDepth];
    }
    
    delete [] m_pppcRDSbacCoder;
    delete [] m_pppcBinCoderCABAC;
  }
  
  // destroy ROM
  destroyROM();
  
  return;
}

Void TEncTop::init()
{
  UInt *aTable4=NULL, *aTable8=NULL;
#if QC_MOD_LCEC
  UInt* aTableLastPosVlcIndex=NULL; 
#endif
  // initialize SPS
  xInitSPS();
  
  // initialize PPS
#if SUB_LCU_DQP
  m_cPPS.setSPS(&m_cSPS);
  m_cPPS.setBDS(&m_cBDS);
#endif
  xInitPPS();

  xInitBDS();


  // initialize processing unit classes
  m_cGOPEncoder.  init( this );
  m_cSliceEncoder.init( this );
  m_cCuEncoder.   init( this );
  
  // initialize transform & quantization class
  m_pcCavlcCoder = getCavlcCoder();
#if !CAVLC_COEF_LRG_BLK
  aTable8 = m_pcCavlcCoder->GetLP8Table();
#endif
  aTable4 = m_pcCavlcCoder->GetLP4Table();
#if QC_MOD_LCEC
  aTableLastPosVlcIndex=m_pcCavlcCoder->GetLastPosVlcIndexTable();
  
  m_cTrQuant.init( g_uiMaxCUWidth, g_uiMaxCUHeight, 1 << m_uiQuadtreeTULog2MaxSize, m_iSymbolMode, aTable4, aTable8, 
    aTableLastPosVlcIndex, m_bUseRDOQ, true );
#else
  m_cTrQuant.init( g_uiMaxCUWidth, g_uiMaxCUHeight, 1 << m_uiQuadtreeTULog2MaxSize, m_iSymbolMode, aTable4, aTable8, m_bUseRDOQ, true );
#endif
  
  // initialize encoder search class
  m_cSearch.init( this, &m_cTrQuant, m_iSearchRange, m_bipredSearchRange, m_iFastSearch, 0, &m_cEntropyCoder, &m_cRdCost, getRDSbacCoder(), getRDGoOnSbacCoder() );

#if MQT_ALF_NPASS
  if(m_bUseALF)
  {
    m_cAdaptiveLoopFilter.setALFEncodePassReduction( m_iALFEncodePassReduction );
  }
#endif

  m_iMaxRefPicNum = 0;
}

// ====================================================================================================================
// Public member functions
// ====================================================================================================================

Void TEncTop::deletePicBuffer()
{
  TComList<TComPic*>::iterator iterPic = m_cListPic.begin();
  Int iSize = Int( m_cListPic.size() );
  
  for ( Int i = 0; i < iSize; i++ )
  {
    TComPic* pcPic = *(iterPic++);
    
    pcPic->destroy();
    delete pcPic;
    pcPic = NULL;
  }
}

/**
 - Application has picture buffer list with size of GOP + 1
 - Picture buffer list acts like as ring buffer
 - End of the list has the latest picture
 .
 \param   bEos                true if end-of-sequence is reached
 \param   pcPicYuvOrg         original YUV picture
 \retval  rcListPicYuvRecOut  list of reconstruction YUV pictures
 \retval  rcListBitstreamOut  list of output bitstreams
 \retval  iNumEncoded         number of encoded pictures
 */
Void TEncTop::encode( bool bEos, TComPicYuv* pcPicYuvOrg, TComList<TComPicYuv*>& rcListPicYuvRecOut, list<AccessUnit>& accessUnitsOut, Int& iNumEncoded )
{
  TComPic* pcPicCurr = NULL;
  
  // get original YUV
  xGetNewPicBuffer( pcPicCurr );
  pcPicYuvOrg->copyToPic( pcPicCurr->getPicYuvOrg() );
  
  if ( m_iPOCLast != 0 && ( m_iNumPicRcvd != m_iGOPSize && m_iGOPSize ) && !bEos )
  {
    iNumEncoded = 0;
    return;
  }
  
  // compress GOP
  m_cGOPEncoder.compressGOP(m_iPOCLast, m_iNumPicRcvd, m_cListPic, rcListPicYuvRecOut, accessUnitsOut);
  
  iNumEncoded         = m_iNumPicRcvd;
  m_iNumPicRcvd       = 0;
  m_uiNumAllPicCoded += iNumEncoded;
  
  if (bEos)
  {
    m_cGOPEncoder.printOutSummary (m_uiNumAllPicCoded);
  }
}

// ====================================================================================================================
// Protected member functions
// ====================================================================================================================

/**
 - Application has picture buffer list with size of GOP + 1
 - Picture buffer list acts like as ring buffer
 - End of the list has the latest picture
 .
 \retval rpcPic obtained picture buffer
 */
Void TEncTop::xGetNewPicBuffer ( TComPic*& rpcPic )
{
  TComSlice::sortPicList(m_cListPic);
  
  if (m_cListPic.size() >= (UInt)(m_iGOPSize + getNumOfReference() + 2) )
  {
    TComList<TComPic*>::iterator iterPic  = m_cListPic.begin();
    Int iSize = Int( m_cListPic.size() );
    for ( Int i = 0; i < iSize; i++ )
    {
      rpcPic = *(++iterPic);
      if(rpcPic->getSlice(0)->isReferenced() == false)
         break;
    }
  }
  else
  {
    rpcPic = new TComPic;
    rpcPic->create( m_iSourceWidth, m_iSourceHeight, g_uiMaxCUWidth, g_uiMaxCUHeight, g_uiMaxCUDepth );
    m_cListPic.pushBack( rpcPic );
  }
  
  rpcPic->setReconMark (false);
  
  m_iPOCLast++;
  m_iNumPicRcvd++;
  
  rpcPic->getSlice(0)->setPOC( m_iPOCLast );
  // mark it should be extended
  rpcPic->getPicYuvRec()->setBorderExtension(false);
}

Void TEncTop::xInitSPS()
{
  m_cSPS.setWidth         ( m_iSourceWidth      );
  m_cSPS.setHeight        ( m_iSourceHeight     );
  m_cSPS.setPad           ( m_aiPad             );
  m_cSPS.setMaxCUWidth    ( g_uiMaxCUWidth      );
  m_cSPS.setMaxCUHeight   ( g_uiMaxCUHeight     );
  m_cSPS.setMaxCUDepth    ( g_uiMaxCUDepth      );
  m_cSPS.setMinTrDepth    ( 0                   );
  m_cSPS.setMaxTrDepth    ( 1                   );

  m_cSPS.setMaxNumberOfReferencePictures(m_uiMaxNumberOfReferencePictures);
  m_cSPS.setMaxNumberOfReorderPictures(m_uiMaxNumberOfReorderPictures);

#if E057_INTRA_PCM
  m_cSPS.setPCMLog2MinSize (m_uiPCMLog2MinSize);
#endif

  m_cSPS.setUseALF        ( m_bUseALF           );
  
  m_cSPS.setQuadtreeTULog2MaxSize( m_uiQuadtreeTULog2MaxSize );
  m_cSPS.setQuadtreeTULog2MinSize( m_uiQuadtreeTULog2MinSize );
  m_cSPS.setQuadtreeTUMaxDepthInter( m_uiQuadtreeTUMaxDepthInter    );
  m_cSPS.setQuadtreeTUMaxDepthIntra( m_uiQuadtreeTUMaxDepthIntra    );
  
  m_cSPS.setUseDQP        ( m_iMaxDeltaQP != 0  );
  m_cSPS.setUseLDC        ( m_bUseLDC           );
  m_cSPS.setUsePAD        ( m_bUsePAD           );
  
  m_cSPS.setUseMRG        ( m_bUseMRG           ); // SOPH:

#if LM_CHROMA 
  m_cSPS.setUseLMChroma   ( m_bUseLMChroma           );  
#endif
  
  m_cSPS.setMaxTrSize   ( 1 << m_uiQuadtreeTULog2MaxSize );
  
#if DCM_COMB_LIST
  m_cSPS.setUseLComb    ( m_bUseLComb           );
  m_cSPS.setLCMod       ( m_bLCMod   );
#endif

  Int i;
#if HHI_AMVP_OFF
  for ( i = 0; i < g_uiMaxCUDepth; i++ )
  {
    m_cSPS.setAMVPMode( i, AM_NONE );
  }
#else
  for ( i = 0; i < g_uiMaxCUDepth; i++ )
  {
    m_cSPS.setAMVPMode( i, AM_EXPL );
  }
#endif
  
  
#if HHI_RMP_SWITCH
  m_cSPS.setUseRMP( m_bUseRMP );
#endif
  
  m_cSPS.setBitDepth    ( g_uiBitDepth        );
  m_cSPS.setBitIncrement( g_uiBitIncrement    );

#if MTK_NONCROSS_INLOOP_FILTER
  m_cSPS.setLFCrossSliceBoundaryFlag( m_bLFCrossSliceBoundaryFlag );
#endif
#if MTK_SAO
  m_cSPS.setUseSAO             ( m_bUseSAO         );
#endif

  if ( m_bTLayering )
  {
    Int iMaxTLayers = 1;
    for ( i = 1; ; i++)
    {
      iMaxTLayers = i;
      if ( (m_iRateGOPSize >> i) == 0 ) 
      {
        break;
      }
    }
  
    m_cSPS.setMaxTLayers( (UInt)iMaxTLayers );

    Bool bTemporalIdNestingFlag = true;
    for ( i = 0; i < m_cSPS.getMaxTLayers()-1; i++ )
    {
      if ( !m_abTLayerSwitchingFlag[i] )
      {
        bTemporalIdNestingFlag = false;
        break;
      }
    }

    m_cSPS.setTemporalIdNestingFlag( bTemporalIdNestingFlag );
  }
  else
  {
    m_cSPS.setMaxTLayers( 1 );
    m_cSPS.setTemporalIdNestingFlag( false );
  }
#if E057_INTRA_PCM && E192_SPS_PCM_BIT_DEPTH_SYNTAX
  m_cSPS.setPCMBitDepthLuma (g_uiPCMBitDepthLuma);
  m_cSPS.setPCMBitDepthChroma (g_uiPCMBitDepthChroma);
#endif
#if E057_INTRA_PCM && E192_SPS_PCM_FILTER_DISABLE_SYNTAX
  m_cSPS.setPCMFilterDisableFlag  ( m_bPCMFilterDisableFlag );
#endif
}

Void TEncTop::xInitPPS()
{
#if CONSTRAINED_INTRA_PRED
  m_cPPS.setConstrainedIntraPred( m_bUseConstrainedIntraPred );
#endif

  if ( m_cSPS.getTemporalIdNestingFlag() ) 
  {
    m_cPPS.setNumTLayerSwitchingFlags( 0 );
    for ( UInt i = 0; i < m_cSPS.getMaxTLayers() - 1; i++ )
    {
      m_cPPS.setTLayerSwitchingFlag( i, true );
    }
  }
  else
  {
    m_cPPS.setNumTLayerSwitchingFlags( m_cSPS.getMaxTLayers() - 1 );
    for ( UInt i = 0; i < m_cPPS.getNumTLayerSwitchingFlags(); i++ )
    {
      m_cPPS.setTLayerSwitchingFlag( i, m_abTLayerSwitchingFlag[i] );
    }
  }  


  Int max_temporal_layers = m_cPPS.getSPS()->getMaxTLayers();
  if(max_temporal_layers > 4)
     m_cPPS.setBitsForTemporalId(3);
  else if(max_temporal_layers > 2)
     m_cPPS.setBitsForTemporalId(2);
  else if(max_temporal_layers > 1)
     m_cPPS.setBitsForTemporalId(1);
  else
     m_cPPS.setBitsForTemporalId(0);

  m_cPPS.setReorderPicturesAllowedFlag(m_cPPS.getSPS()->getMaxNumberOfReorderPictures() > 0);

#if SUB_LCU_DQP
  if( m_cPPS.getSPS()->getUseDQP() )
  {
    m_cPPS.setMaxCuDQPDepth( m_iMaxCuDQPDepth );
    m_cPPS.setMinCuDQPSize( m_cPPS.getSPS()->getMaxCUWidth() >> ( m_cPPS.getMaxCuDQPDepth()) );
  }
  else
  {
    m_cPPS.setMaxCuDQPDepth( 0 );
    m_cPPS.setMinCuDQPSize( m_cPPS.getSPS()->getMaxCUWidth() >> ( m_cPPS.getMaxCuDQPDepth()) );
  }
#endif
}

Void TEncTop::xInitBDS()
{
   TComBufferDescription*      pBD;

   // In this function 
   // a number of Buffer Descriptions
   // are defined for differnt coding structures.
   // This is the place 
   // where you need to do changes in
   // order to try different Buffer Descriptions.
   // In a future implementation the 
   // Buffer Descriptions will be 
   // configured directly from the config file.

   // Here we check what BD is appropriate
   UInt bd_structure;
   if(getIntraPeriod() == 1) // Intra-only
   {
      bd_structure = 0;
   }
   else if(this->getPPS()->getSPS()->getUseLDC()) // Low Delay
   {
#if ERICSSON_USE_SPECIALIZED_BD_FOR_LD == 1
      bd_structure = 1;      //this uses the closest 2 frames and 2 high quality
#else
      bd_structure = 2;    //this uses the closest 4 frames
#endif
   }
   else
   {
      bd_structure = 3;
   }

   if(bd_structure == 0)
   {
      // This is a Buffer Description Set for
      // Intra Only Coding structure with 
      // 1 Buffer Description containing 
      // 0 reference pictures
      m_cBDS.create(1);
      pBD = m_cBDS.getBufferDescription(0);
      pBD->create(0); 
      pBD->setNumberOfNegativePictures(0); 
      pBD->setNumberOfPositivePictures(0); 
      return;
   }

   if(bd_structure == 1)
   {
      // This is a Buffer Description Set for
      // a Low Delay Coding structure with 
      // 4 Buffer Descriptions containing 
      // 4 reference pictures each
      // the two closest + two of high quality
      m_cBDS.create(4);

      //BD 0
      pBD = m_cBDS.getBufferDescription(0);
      pBD->create(4);      
      pBD->setNumberOfNegativePictures(4);
      pBD->setDeltaPOC(0, -1);
      pBD->setTemporalId(0, 2);
      pBD->setDeltaPOC(1, -2);
      pBD->setTemporalId(1, 1);
      pBD->setDeltaPOC(2, -4);
      pBD->setTemporalId(2, 0);
      pBD->setDeltaPOC(3, -8);
      pBD->setTemporalId(3, 0);

      //BD 1
      pBD = m_cBDS.getBufferDescription(1);
      pBD->create(4);   
      pBD->setNumberOfNegativePictures(4);   
      pBD->setDeltaPOC(0, -1);
      pBD->setTemporalId(0, 0);
      pBD->setDeltaPOC(1, -2);
      pBD->setTemporalId(1, 2);
      pBD->setDeltaPOC(2, -5);
      pBD->setTemporalId(2, 0);
      pBD->setDeltaPOC(3, -9);
      pBD->setTemporalId(3, 0);

      //BD 2
      pBD = m_cBDS.getBufferDescription(2);
      pBD->create(4);     
      pBD->setNumberOfNegativePictures(4); 
      pBD->setDeltaPOC(0, -1);
      pBD->setTemporalId(0, 2);
      pBD->setDeltaPOC(1, -2);
      pBD->setTemporalId(1, 0);
      pBD->setDeltaPOC(2, -6);
      pBD->setTemporalId(2, 0);
      pBD->setDeltaPOC(3, -10);
      pBD->setTemporalId(3, 0);

      //BD 3
      pBD = m_cBDS.getBufferDescription(3);
      pBD->create(4);    
      pBD->setNumberOfNegativePictures(4);  
      pBD->setDeltaPOC(0, -1);
      pBD->setTemporalId(0, 1);
      pBD->setDeltaPOC(1, -2);
      pBD->setTemporalId(1, 2);
      pBD->setDeltaPOC(2, -3);
      pBD->setTemporalId(2, 0);
      pBD->setDeltaPOC(3, -7);
      pBD->setTemporalId(3, 0);
   }

   
   if(bd_structure == 2)
   {
      // This is a Buffer Description Set for
      // a Low Delay Coding structure with 
      // 1 Buffer Description containing 
      // the 4 closest reference pictures
      // this structure assumes 
      // TLayeringBasedOnCodingStruct = 0 
      // (as in common conditions)
      m_cBDS.create(1);

      //BD 0
      pBD = m_cBDS.getBufferDescription(0);
      pBD->create(4);   
      pBD->setNumberOfNegativePictures(4);   
      pBD->setDeltaPOC(0, -1);
      pBD->setTemporalId(0, 0);
      pBD->setDeltaPOC(1, -2);
      pBD->setTemporalId(1, 0);
      pBD->setDeltaPOC(2, -3);
      pBD->setTemporalId(2, 0);
      pBD->setDeltaPOC(3, -4);
      pBD->setTemporalId(3, 0);
   }

   //Random Access
   if(bd_structure == 3)
   {
      // This is a Buffer Description Set for
      // a Random Access Coding structure with 
      // 13 Buffer Descriptions containing 
      // up to 6 reference pictures
      m_cBDS.create(13);

      // BD 0
      pBD = m_cBDS.getBufferDescription(0);
      pBD->create(4);   
      pBD->setNumberOfNegativePictures(4);    
      pBD->setDeltaPOC(0, -8);
      pBD->setTemporalId(0, 0);
      pBD->setDeltaPOC(1, -10);
      pBD->setTemporalId(1, 2);
      pBD->setDeltaPOC(2, -12);
      pBD->setTemporalId(2, 1);
      pBD->setDeltaPOC(3, -14);
      pBD->setTemporalId(3, 2);  
      pBD->setNumberOfPositivePictures(0); 

      // BD 4
      pBD = m_cBDS.getBufferDescription(4);
      pBD->create(3);  
      pBD->setNumberOfNegativePictures(2); 
      pBD->setDeltaPOC(0, -4);
      pBD->setTemporalId(0, 0);
      pBD->setDeltaPOC(1, -6);
      pBD->setTemporalId(1, 2);  
      pBD->setNumberOfPositivePictures(1);    
      pBD->setDeltaPOC(2, 4);
      pBD->setTemporalId(2, 0);

      //BD 2
      pBD = m_cBDS.getBufferDescription(2);
      pBD->create(4);      
      pBD->setNumberOfNegativePictures(2); 
      pBD->setDeltaPOC(0, -2);
      pBD->setTemporalId(0, 0);
      pBD->setDeltaPOC(1, -4);
      pBD->setTemporalId(1, 1);  
      pBD->setNumberOfPositivePictures(2); 
      pBD->setDeltaPOC(2, 2);
      pBD->setTemporalId(2, 1);
      pBD->setDeltaPOC(3, 6);
      pBD->setTemporalId(3, 0);

      //BD 6
      pBD = m_cBDS.getBufferDescription(6);
      pBD->create(5);      
      pBD->setNumberOfNegativePictures(4);  
      pBD->setDeltaPOC(0, -2);
      pBD->setTemporalId(0, 1);
      pBD->setDeltaPOC(1, -4);
      pBD->setTemporalId(1, 2);
      pBD->setDeltaPOC(2, -6);
      pBD->setTemporalId(2, 0);
      pBD->setDeltaPOC(3, -8);
      pBD->setTemporalId(3, 2); 
      pBD->setNumberOfPositivePictures(1); 
      pBD->setDeltaPOC(4, 2);
      pBD->setTemporalId(4, 0);

      //BD 1
      pBD = m_cBDS.getBufferDescription(1);
      pBD->create(6);      
      pBD->setNumberOfNegativePictures(2); 
      pBD->setDeltaPOC(0, -1);
      pBD->setTemporalId(0, 0);
      pBD->setDeltaPOC(1, -3);
      pBD->setTemporalId(1, 2);  
      pBD->setNumberOfPositivePictures(4); 
      pBD->setDeltaPOC(2, 1);
      pBD->setTemporalId(2, 2);
      pBD->setDeltaPOC(3, 3);
      pBD->setTemporalId(3, 1);
      pBD->setDeltaPOC(4, 5);
      pBD->setTemporalId(4, 2);
      pBD->setDeltaPOC(5, 7);
      pBD->setTemporalId(5, 0);

      //BD 3
      pBD = m_cBDS.getBufferDescription(3);  
      pBD->create(5);      
      pBD->setNumberOfNegativePictures(2); 
      pBD->setDeltaPOC(0, -1);
      pBD->setTemporalId(0, 2);
      pBD->setDeltaPOC(1, -3);
      pBD->setTemporalId(1, 0);  
      pBD->setNumberOfPositivePictures(3); 
      pBD->setDeltaPOC(2, 1);
      pBD->setTemporalId(2, 1);
      pBD->setDeltaPOC(3, 3);
      pBD->setTemporalId(3, 2);
      pBD->setDeltaPOC(4, 5);
      pBD->setTemporalId(4, 0);

      //BD 5
      pBD = m_cBDS.getBufferDescription(5);
      pBD->create(4);      
      pBD->setNumberOfNegativePictures(2); 
      pBD->setDeltaPOC(0, -1);
      pBD->setTemporalId(0, 1);
      pBD->setDeltaPOC(1, -3);
      pBD->setTemporalId(1, 2);  
      pBD->setNumberOfPositivePictures(2); 
      pBD->setDeltaPOC(2, 1);
      pBD->setTemporalId(2, 2);
      pBD->setDeltaPOC(3, 3);
      pBD->setTemporalId(3, 0);

      //BD 7
      pBD = m_cBDS.getBufferDescription(7);
      pBD->create(4);      
      pBD->setNumberOfNegativePictures(3);   
      pBD->setDeltaPOC(0, -1);
      pBD->setTemporalId(0, 2);
      pBD->setDeltaPOC(1, -3);
      pBD->setTemporalId(1, 1);
      pBD->setDeltaPOC(2, -5);
      pBD->setTemporalId(2, 2);
      pBD->setNumberOfPositivePictures(1); 
      pBD->setDeltaPOC(3, 1);
      pBD->setTemporalId(3, 0);

      //The following Buffer Descriptions 
      //are used after an IDR or CDR

      //BD 8
      pBD = m_cBDS.getBufferDescription(8);
      pBD->create(1);      
      pBD->setNumberOfNegativePictures(1);  
      pBD->setDeltaPOC(0, -8);
      pBD->setTemporalId(0, 0); 
      pBD->setNumberOfPositivePictures(0); 

      //BD 9
      pBD = m_cBDS.getBufferDescription(9);
      pBD->create(2);      
      pBD->setNumberOfNegativePictures(1);  
      pBD->setDeltaPOC(0, -4);
      pBD->setTemporalId(0, 0); 
      pBD->setNumberOfPositivePictures(1); 
      pBD->setDeltaPOC(1, 4);
      pBD->setTemporalId(1, 0);

      //BD 10
      pBD = m_cBDS.getBufferDescription(10);
      pBD->create(3);      
      pBD->setNumberOfNegativePictures(1);  
      pBD->setDeltaPOC(0, -2);
      pBD->setTemporalId(0, 0); 
      pBD->setNumberOfPositivePictures(2); 
      pBD->setDeltaPOC(1, 2);
      pBD->setTemporalId(1, 1);
      pBD->setDeltaPOC(2, 6);
      pBD->setTemporalId(2, 0);

      //BD 11
      pBD = m_cBDS.getBufferDescription(11);
      pBD->create(4);      
      pBD->setNumberOfNegativePictures(3); 
      pBD->setDeltaPOC(0, -2);
      pBD->setTemporalId(0, 1);
      pBD->setDeltaPOC(1, -4);
      pBD->setTemporalId(1, 2);
      pBD->setDeltaPOC(2, -6);
      pBD->setTemporalId(2, 0);  
      pBD->setNumberOfPositivePictures(1); 
      pBD->setDeltaPOC(3, 2);
      pBD->setTemporalId(3, 0);

      //BD 12
      pBD = m_cBDS.getBufferDescription(12);
      pBD->create(5);   
      pBD->setNumberOfNegativePictures(1);   
      pBD->setDeltaPOC(0, -1);
      pBD->setTemporalId(0, 0);
      pBD->setNumberOfPositivePictures(4);    
      pBD->setDeltaPOC(1, 1);
      pBD->setTemporalId(1, 2);
      pBD->setDeltaPOC(2, 3);
      pBD->setTemporalId(2, 1);
      pBD->setDeltaPOC(3, 5);
      pBD->setTemporalId(3, 2); 
      pBD->setDeltaPOC(4, 7);
      pBD->setTemporalId(4, 0);  
   }
}

Void TEncTop::SelectBufferDescripitonFromPOC(TComSlice* pcSlice, UInt uiPOCCurr, UInt uiPOCCDR)
{

   // This is a function that 
   // decides what Buffer Description to use 
   // for a specific picture (with POC = uiPOCCurr)
   // In a future implementation
   // this will be controlled by 
   // config file parameters

   if (getIntraPeriod() == 1)
      pcSlice->setBDidx(0);
   else if (pcSlice->getPPS()->getSPS()->getUseLDC())
   {
#if ERICSSON_USE_SPECIALIZED_BD_FOR_LD == 1
      if(uiPOCCurr > 6)
      {
         pcSlice->setBDidx(uiPOCCurr%(4));
      }
      else // explicit Buffer Description for the first 7 pictures
      {
         TComBufferDescription*      pBD;
         pBD = pcSlice->getLocalBD();
         if(uiPOCCurr == 0)
         {
            pBD->create(0);
            pBD->setNumberOfNegativePictures(0); 
         }
         else if(uiPOCCurr == 1)
         {
            pBD->create(1);
            pBD->setNumberOfNegativePictures(1);  
            pBD->setDeltaPOC(0, -1);
            pBD->setTemporalId(0, 0);
         }       
         else if(uiPOCCurr == 2)
         {
            pBD->create(2);
            pBD->setNumberOfNegativePictures(2); 
            pBD->setDeltaPOC(0, -1);
            pBD->setTemporalId(0, 2);
            pBD->setDeltaPOC(1, -2);
            pBD->setTemporalId(1, 0);
         }     
         else if(uiPOCCurr == 3)
         {
            pBD->create(3);
            pBD->setNumberOfNegativePictures(3); 
            pBD->setDeltaPOC(0, -1);
            pBD->setTemporalId(0, 1);
            pBD->setDeltaPOC(1, -2);
            pBD->setTemporalId(1, 2);
            pBD->setDeltaPOC(2, -3);
            pBD->setTemporalId(2, 0);
         }     
         else if(uiPOCCurr == 4)
         {
            pBD->create(4);
            pBD->setNumberOfNegativePictures(4); 
            pBD->setDeltaPOC(0, -1);
            pBD->setTemporalId(0, 2);
            pBD->setDeltaPOC(1, -2);
            pBD->setTemporalId(1, 1);
            pBD->setDeltaPOC(2, -3);
            pBD->setTemporalId(2, 2);
            pBD->setDeltaPOC(3, -4);
            pBD->setTemporalId(3, 0);
         }        
         else if(uiPOCCurr == 5)
         {
            pBD->create(4);
            pBD->setNumberOfNegativePictures(4); 
            pBD->setDeltaPOC(0, -1);
            pBD->setTemporalId(0, 0);
            pBD->setDeltaPOC(1, -2);
            pBD->setTemporalId(1, 2);
            pBD->setDeltaPOC(2, -3);
            pBD->setTemporalId(2, 1);
            pBD->setDeltaPOC(3, -5);
            pBD->setTemporalId(3, 0);
         }           
         else if(uiPOCCurr == 6)
         {
            pBD->create(4);
            pBD->setNumberOfNegativePictures(4); 
            pBD->setDeltaPOC(0, -1);
            pBD->setTemporalId(0, 2);
            pBD->setDeltaPOC(1, -2);
            pBD->setTemporalId(1, 0);
            pBD->setDeltaPOC(2, -4);
            pBD->setTemporalId(2, 1);
            pBD->setDeltaPOC(3, -6);
            pBD->setTemporalId(3, 0);
         }      
         pcSlice->setBD(pBD);
         pcSlice->setBDidx(-1);
      }
#else
      if(uiPOCCurr > 3)
      {
         pcSlice->setBDidx(0);
      }
      else // explicit Buffer Description for the first 3 pictures
      {
         TComBufferDescription*      pBD;
         pBD = pcSlice->getLocalBD();
         if(uiPOCCurr == 0)
         {
            pBD->create(0);
         }
         if(uiPOCCurr == 1)
         {
            pBD->create(1);
            pBD->setNumberOfNegativePictures(1); 
            pBD->setDeltaPOC(0, -1);
            pBD->setTemporalId(0, 0);
         }       
         if(uiPOCCurr == 2)
         {
            pBD->create(2);
            pBD->setNumberOfNegativePictures(2); 
            pBD->setDeltaPOC(0, -1);
            pBD->setTemporalId(0, 0);
            pBD->setDeltaPOC(1, -2);
            pBD->setTemporalId(1, 0);
         }     
         if(uiPOCCurr == 3)
         {
            pBD->create(3);
            pBD->setNumberOfNegativePictures(3); 
            pBD->setDeltaPOC(0, -1);
            pBD->setTemporalId(0, 0);
            pBD->setDeltaPOC(1, -2);
            pBD->setTemporalId(1, 0);
            pBD->setDeltaPOC(2, -3);
            pBD->setTemporalId(2, 0);
         }   
         pcSlice->setBD(pBD);
         pcSlice->setBDidx(-1);
      }
#endif
   }
   else 
   {
      Int POCdiff = uiPOCCurr - uiPOCCDR;
      if(uiPOCCurr == 0)
      {
         TComBufferDescription*      pBD;
         pBD = pcSlice->getLocalBD();
         pBD->create(0);
         pBD->setNumberOfNegativePictures(0); 
         pBD->setNumberOfPositivePictures(0);
         pcSlice->setBD(pBD);
         pcSlice->setBDidx(-1);
      }
      else if (POCdiff == 8)
         pcSlice->setBDidx(8);
      else if (POCdiff == 4)
         pcSlice->setBDidx(9);
      else if (POCdiff == 2)
         pcSlice->setBDidx(10);
      else if (POCdiff == 6)
         pcSlice->setBDidx(11);
      else if (POCdiff == 1)
         pcSlice->setBDidx(12);
      else 
         pcSlice->setBDidx(uiPOCCurr%(8));
   }

   if(pcSlice->getBDidx() >= 0)
      pcSlice->setBD(getBDS()->getBufferDescription(pcSlice->getBDidx()));

   pcSlice->getBD()->setNumberOfPictures(pcSlice->getBD()->getNumberOfNegativePictures()+pcSlice->getBD()->getNumberOfPositivePictures());

}
